import nvdiffrast.torch as dr
import torch

from easyhec.utils import utils_3d


class NVDiffrastRenderer:
    def __init__(self, height: int, width: int):
        self.H, self.W = height, width
        self.resolution = (height, width)
        blender2opencv = (
            torch.tensor([[1, 0, 0, 0], [0, -1, 0, 0], [0, 0, -1, 0], [0, 0, 0, 1]])
            .float()
            .cuda()
        )
        self.opencv2blender = torch.inverse(blender2opencv)
        self.glctx = dr.RasterizeCudaContext()

    def render_mask(
        self,
        verts: torch.Tensor,
        faces: torch.Tensor,
        intrinsic: torch.Tensor,
        object_pose: torch.Tensor,
        anti_aliasing: bool = True,
    ) -> torch.Tensor:
        """
        Differentiable rendering of given vertices and faces given the pose of the object and intrinsics of the camera

        Parameters:
            verts (torch.Tensor, shape (N, 3)): vertices of the object
            faces (torch.Tensor, shape (M, 3)): faces of the object
            intrinsic (torch.Tensor, shape (3, 3)): intrinsic matrix of the camera
            object_pose (torch.Tensor, shape (4, 4)): pose of the object
            anti_aliasing (bool): Default is True. If True, will use antialiasing.

        """
        proj = utils_3d.K_to_projection(intrinsic, self.H, self.W)
        pose = self.opencv2blender @ object_pose
        pos_clip = utils_3d.transform_pos(proj @ pose, verts)
        rast_out, _ = dr.rasterize(
            self.glctx, pos_clip, faces, resolution=self.resolution
        )
        if anti_aliasing:
            vtx_color = torch.ones(verts.shape, dtype=torch.float, device=verts.device)
            color, _ = dr.interpolate(vtx_color[None, ...], rast_out, faces)
            color = dr.antialias(color, rast_out, pos_clip, faces)
            mask = color[0, :, :, 0]
        else:
            mask = rast_out[0, :, :, 2] > 0
        mask = torch.flip(mask, dims=[0])
        return mask



"""
Tools for interactive segmentation
"""

import sys

import matplotlib.pyplot as plt
import numpy as np
import torch
from matplotlib import pyplot as plt


class ImageRenderer:
    def __init__(self, wait_for_button_press=True):
        """
        Create a very light-weight image renderer.

        Args:
            wait_for_button_press (bool): If True, each call to this renderer will pause the process until the user presses any key.
            event_handler: Code to run given an event / button press. If None the default is mapping 'escape' and 'q' to sys.exit(0)
        """
        self._image = None
        self.last_event = None
        self.wait_for_button_press = wait_for_button_press
        self.pressed_keys = set()

    def key_press_handler(self, event):
        self.last_event = event
        self.pressed_keys.add(event.key)
        if event.key in ["q", "escape"]:
            sys.exit(0)

    def key_release_handler(self, event):
        if event.key in self.pressed_keys:
            self.pressed_keys.remove(event.key)

    def __call__(self, buffer):
        if not self._image:
            plt.ion()
            self.fig, self.ax = plt.subplots()
            self._image = self.ax.imshow(buffer, animated=True)
            self.fig.canvas.mpl_connect("key_press_event", self.key_press_handler)
            self.fig.canvas.mpl_connect("key_release_event", self.key_release_handler)
        else:
            self._image.set_data(buffer)
        if self.wait_for_button_press:
            plt.waitforbuttonpress()
        else:
            self.fig.canvas.draw_idle()
            self.fig.canvas.flush_events()
        plt.draw()

    def __del__(self):
        self.close()

    def close(self):
        plt.ioff()
        plt.close()


class InteractiveSegmentation:
    """
    Interactive segmentation tool. Opens a window from which you can click to record pixel positions.
    """

    def __init__(
        self,
        segmentation_model: str = "sam2",
        segmentation_model_cfg: dict = dict(
            checkpoint="sam2/checkpoints/sam2.1_hiera_large.pt",
            model_cfg="sam2/configs/sam2.1/sam2.1_hiera_l.yaml",
        ),
    ):
        """
        Args:
            segmentation_model: The segmentation model to use. Currently only "sam2" is supported.
        """
        self.segmentation_model = segmentation_model
        if self.segmentation_model == "sam2":
            from sam2.build_sam import build_sam2
            from sam2.sam2_image_predictor import SAM2ImagePredictor

            model_cfg = segmentation_model_cfg["model_cfg"]
            checkpoint = segmentation_model_cfg["checkpoint"]
            self.predictor = SAM2ImagePredictor(build_sam2(model_cfg, checkpoint))
        else:
            raise ValueError(
                f"Segmentation model {self.segmentation_model} not supported"
            )

    def get_segmentation(self, images: np.ndarray):
        """
        Get segmentation from a list of imagees. Opens a window from which you can click to record points of the object to segment out.

        There are a few other options that let the user e.g. redo the segmentation, redo the points etc., see the terminal output for help
        """
        renderer = ImageRenderer(wait_for_button_press=False)
        state = "annotation"
        current_image_idx = 0
        masks = []
        annotation_objs = []
        clicked_points = []

        def print_help_message():
            if state == "annotation":
                print(
                    f"Currently annotating image {current_image_idx+1}/{len(images)}. Click to add a point of what to segment, right click to add a negative point of what not to segment. Press 't' when done. Press 'r' to clear the current point annotation and redo the points"
                )
            elif state == "segmentation":
                print(
                    f"Currently showing the predicted segmentation for image {current_image_idx+1}/{len(images)}. Press 't' to move on to the next image. Press 'e' to delete this segmentation and edit the existing annotation points. Press 'r' to delete this segmentation and re-annotate the points for this image."
                )

        def onclick(event):
            nonlocal annotation_objs, clicked_points
            if event.xdata is not None and event.ydata is not None:
                x, y = int(event.xdata), int(event.ydata)
                if event.button == 3:
                    clicked_points.append((x, y, 0))
                    annotation_objs.append(plt.plot(x, y, "ro")[0])
                else:
                    if x < 0 or x >= image.shape[1] or y < 0 or y >= image.shape[0]:
                        return
                    clicked_points.append((x, y, 1))
                    annotation_objs.append(plt.plot(x, y, "go")[0])

        def clear_drawn_points():
            nonlocal annotation_objs
            for x in annotation_objs:
                x.remove()
            annotation_objs = []

        renderer(images[0])
        renderer.ax.axis("off")
        cid = None
        print(
            f"Starting annotation process for {len(images)} images. Press 't' to finish annotation, 'r' to redo annotation. Press 'h' for help."
        )
        print("--------------------------------")
        print_help_message()
        while current_image_idx < len(images):
            image = images[current_image_idx].copy()
            key = renderer.last_event.key if renderer.last_event is not None else None
            if renderer.last_event is not None:
                renderer.last_event = None
            if key == "q":
                renderer.close()
                return None
            if key == "h":
                print_help_message()
            if state == "annotation":
                cid = renderer.fig.canvas.mpl_connect("button_press_event", onclick)
                renderer.ax.set_title(
                    "Click on the image to record annotation points for segmentation"
                )
                renderer(image)

                if key == "t":
                    if self.segmentation_model == "sam2":
                        clicked_points_np = np.array(clicked_points)
                        input_label = clicked_points_np[:, 2]
                        input_point = clicked_points_np[:, :2]
                        with torch.inference_mode(), torch.autocast(
                            "cuda", dtype=torch.bfloat16
                        ):
                            self.predictor.set_image(image)
                            mask, _, _ = self.predictor.predict(
                                input_point, input_label, multimask_output=False
                            )
                            mask = mask[0]
                    state = "segmentation"
                    clear_drawn_points()
                elif key == "r":
                    clear_drawn_points()
                    clicked_points = []
                    print("Cleared previous points")
            elif state == "segmentation":
                renderer.fig.canvas.mpl_disconnect(cid)
                masked_image = image.copy()
                mask_color = np.array([30, 144, 255])
                mask_overlay = mask.astype(float).reshape(
                    image.shape[0], image.shape[1], 1
                ) * mask_color.reshape(1, 1, -1)
                masked_image = mask_overlay * 0.6 + masked_image * 0.4
                masked_image[mask == 0] = image[mask == 0]
                renderer.ax.set_title("Check the segmentation quality")
                renderer(masked_image.astype(np.uint8))
                if key == "t":
                    masks.append(mask)
                    current_image_idx += 1
                    state = "annotation"
                    clear_drawn_points()
                    clicked_points = []
                    if current_image_idx < len(images):
                        print_help_message()
                elif key == "e":
                    state = "annotation"
                    # redraw existing points since they got removed to show the segmentation image
                    for x in annotation_objs:
                        x.remove()
                    annotation_objs = []
                    for pos in clicked_points:
                        annotation_objs.append(
                            renderer.ax.plot(
                                pos[0], pos[1], "ro" if pos[2] == 0 else "go"
                            )[0]
                        )
                elif key == "r":
                    state = "annotation"
                    clicked_points = []
                    clear_drawn_points()
                    print("Cleared previous points")
        renderer.close()
        return np.stack(masks)



import numpy as np


def ros2opencv(transformation: np.ndarray):
    """
    Convert a ROS convention to opencv convention transformation.
    """
    ros2opencv = np.array([[0, 0, 1, 0], [-1, 0, 0, 0], [0, -1, 0, 0], [0, 0, 0, 1]]).T
    opencv_transformation = ros2opencv @ np.linalg.inv(transformation)
    return opencv_transformation


def opencv2ros(transformation: np.ndarray):
    """
    Convert an opencv convention transformation to a ROS convention transformation.
    """
    ros2opencv = np.array([[0, 0, 1, 0], [-1, 0, 0, 0], [0, -1, 0, 0], [0, 0, 0, 1]]).T
    ros_transformation = np.linalg.inv(ros2opencv.T @ transformation)
    return ros_transformation



import cv2
import numpy as np
import torch
import trimesh

from easyhec.utils.pytorch3d_se3 import (_get_se3_V_input, _se3_V_matrix,
                                         se3_exp_map)
from easyhec.utils.pytorch3d_se3 import se3_log_map as se3_log_map_pytorch3d


def merge_meshes(meshes: list[trimesh.Trimesh]):
    n, vs, fs = 0, [], []
    for mesh in meshes:
        v, f = mesh.vertices, mesh.faces
        vs.append(v)
        fs.append(f + n)
        n = n + v.shape[0]
    if n:
        return trimesh.Trimesh(np.vstack(vs), np.vstack(fs))
    else:
        return None


def K_to_projection(K, H, W, n=0.001, f=10.0):
    fu, fv, cu, cv = K[0, 0], K[1, 1], K[0, 2], K[1, 2]
    proj = (
        torch.tensor(
            [
                [2 * fu / W, 0, -2 * cu / W + 1, 0],
                [0, 2 * fv / H, 2 * cv / H - 1, 0],
                [0, 0, -(f + n) / (f - n), -2 * f * n / (f - n)],
                [0, 0, -1, 0],
            ]
        )
        .cuda()
        .float()
    )
    return proj


def transform_pos(mtx, pos):
    t_mtx = torch.from_numpy(mtx).cuda() if isinstance(mtx, np.ndarray) else mtx
    # (x,y,z) -> (x,y,z,1)
    posw = torch.cat([pos, torch.ones([pos.shape[0], 1]).cuda()], axis=1)
    return torch.matmul(posw, t_mtx.t())[None, ...]


def se3_log_map(
    transform: torch.Tensor,
    eps: float = 1e-4,
    cos_bound: float = 1e-4,
    backend=None,
    test_acc=True,
):
    if backend is None:
        backend = "pytorch3d"
    if backend == "pytorch3d":
        dof6 = se3_log_map_pytorch3d(transform, eps, cos_bound)
    elif backend == "opencv":
        # from pytorch3d.common.compat import solve
        log_rotation = []
        for tsfm in transform:
            cv2_rot = -cv2.Rodrigues(tsfm[:3, :3].detach().cpu().numpy().astype(float))[
                0
            ]
            log_rotation.append(
                torch.from_numpy(cv2_rot.reshape(-1)).to(transform.device).float()
            )
        log_rotation = torch.stack(log_rotation, dim=0)
        T = transform[:, 3, :3]
        V = _se3_V_matrix(*_get_se3_V_input(log_rotation), eps=eps)
        log_translation = torch.linalg.solve(V, T[:, :, None])[:, :, 0]
        dof6 = torch.cat((log_translation, log_rotation), dim=1)
    else:
        raise NotImplementedError()
    if test_acc:
        err = (se3_exp_map(dof6) - transform).abs().max()
        if err > 0.1:
            raise RuntimeError()
    return dof6



import logging
import time
from collections import defaultdict
from dataclasses import dataclass
from pathlib import Path
from typing import Optional

import cv2
import gymnasium as gym
import numpy as np
import pyrealsense2 as rs
import torch
import trimesh
import tyro
from lerobot.cameras.realsense import RealSenseCamera
from lerobot.cameras.realsense.configuration_realsense import \
    RealSenseCameraConfig
from lerobot.motors.motors_bus import MotorNormMode
from lerobot.robots.robot import Robot
from lerobot.robots.so100_follower.config_so100_follower import \
    SO100FollowerConfig
from lerobot.robots.so100_follower.so100_follower import SO100Follower
from lerobot.robots.utils import make_robot_from_config
from transforms3d.euler import euler2mat
from urchin import URDF

from easyhec.examples.real.base import Args
from easyhec.optim.optimize import optimize
from easyhec.segmentation.interactive import InteractiveSegmentation
from easyhec.utils import visualization
from easyhec.utils.camera_conversions import opencv2ros, ros2opencv
from easyhec.utils.utils_3d import merge_meshes


@dataclass
class SO100Args(Args):
    """Calibrate a (realsense) camera with LeRobot SO100. Note that this script might not work with your particular realsense camera, modify as needed. Other cameras can work if you modify the code to get the camera intrinsics and a single color image from the camera. Results are saved to {output_dir} and organized by the camera name specified in the robot config. Currently only supports off-hand cameras
    
    For your own usage you may have a different camera setup, robot, calibration offsets etc., so we recommend you to copy this file at https://github.com/stonet2000/simple-easyhec/blob/main/easyhec/examples/real/so100.py. 
    
    Before usage make sure to calibrate the robot's motors according to the LeRobot tutorial and look for all comments that start with "CHECK:" which highlight the following:

    1. Check the robot config and make sure the correct camera is used. The default script is for a single realsense camera labelled as "base_camera".
    2. Check and modify the CALIBRATION_OFFSET dictionary to match your own robot's calibration offsets. This is extremely important to tune and is necessary since the 0 degree position of the joints in the real world when calibrated with LeRobot currently do not match the 0 degree position when rendered/simulated.
    3. Modify the initial extrinsic guess if the optimization process fails to converge to a good solution. To save time you can also turn on --use-previous-captures to skip the data collection process if already done once.

    Note that LeRobot SO100 motor calibration is done by moving most joints from one end to another. Make sure to move the joints are far as possible during the LeRobot tutorial on caibration for best results.

    """
    output_dir: str = "results/so100"
    use_previous_captures: bool = False
    """If True, will use the previous collected images and robot segmentations if they exist which can save you time. Otherwise, will prompt you to generate a new segmentation mask. This is useful if you find the initial extrinsic guess is not good enough and simply want to refine that and want to skip the segmentation process."""

    robot_id: Optional[str] = None
    """LeRobot robot ID. If provided will control that robot and will save results to {output_dir}/{robot_id}"""
    realsense_camera_serial_id: str = "146322070293"
    """Realsense camera serial ID."""

# CHECK: This is extrememly important to tune. Run this script with --help for an explanation.
CALIBRATION_OFFSET = {
    "shoulder_pan": 0,
    "shoulder_lift": 0,
    "elbow_flex": 0,
    "wrist_flex": 0,
    "wrist_roll": 0,
    "gripper": 0,
}

# For the author's SO100 they used this calibration offset. Yours might be different
CALIBRATION_OFFSET = {
    "shoulder_pan": -3,
    "shoulder_lift": -3,
    "elbow_flex": 5,
    "wrist_flex": 5,
    "wrist_roll": 0,
    "gripper": 0,
}

# CHECK: Check that the created robot config matches the one you wish to use and sets up the port, cameras etc. correctly.
def create_real_robot(uid: str = "so100", robot_id: Optional[str] = None, realsense_serial_number: str = "146322070293") -> Robot:
    """Wrapper function to map string UIDS to real robot configurations. Primarily for saving a bit of code for users when they fork the repository. They can just edit the camera, id etc. settings in this one file."""
    if uid == "so100":
        robot_config = SO100FollowerConfig(
            port="/dev/ttyACM0",
            use_degrees=True,
            # for phone camera users you can use the commented out setting below
            # cameras={
            #     "base_camera": OpenCVCameraConfig(camera_index=1, fps=30, width=640, height=480)
            # }
            # for intel realsense camera users you need to modify the serial number or name for your own hardware
            cameras={
                "base_camera": RealSenseCameraConfig(serial_number_or_name=realsense_serial_number, fps=30, width=1280, height=720)
            },
            id=robot_id,
        )
        real_robot = make_robot_from_config(robot_config)
        return real_robot


def main(args: SO100Args):
    user_tuned_calibration_offset = False
    for k in CALIBRATION_OFFSET.keys():
        if CALIBRATION_OFFSET[k] != 0:
            user_tuned_calibration_offset = True
            break
    if not user_tuned_calibration_offset:
        logging.warning("The calibration offset for sim2real/real2sim is not tuned!! Unless you are absolutely sure you will most likely get poor results.")

    robot_id = "default" if args.robot_id is None else args.robot_id
    robot: SO100Follower = create_real_robot("so100", robot_id=args.robot_id, realsense_serial_number=args.realsense_camera_serial_id)
    robot.bus.motors["gripper"].norm_mode = MotorNormMode.DEGREES
    robot.connect()

    cameras_ft = robot._cameras_ft
    print(f"Found {len(cameras_ft)} cameras to calibrate")
    for k in cameras_ft.keys():
        (Path(args.output_dir) / robot_id / k).mkdir(parents=True, exist_ok=True)
    
    ### Make an initial guess for the extrinsic for each camera ###
    # CHECK: Double check this initial extrinsic guess is roughly close to the real world.
    initial_extrinsic_guesses = dict()
    for k in cameras_ft.keys():
        initial_extrinsic_guess = np.eye(4)

        # the guess says we are at position xyz=[-0.4, 0.0, 0.4] and angle the camerea downwards by np.pi / 4 radians  or 45 degrees
        # note that this convention is more natural for robotics (follows the typical convention for ROS and various simulators), where +Z is moving up towards the sky, +Y is to the left, +X is forward
        initial_extrinsic_guess[:3, :3] = euler2mat(0, np.pi / 4, -np.pi / 5)
        initial_extrinsic_guess[:3, 3] = np.array([-0.4, 0.1, 0.5])
        initial_extrinsic_guess = ros2opencv(initial_extrinsic_guess)

        initial_extrinsic_guesses[k] = initial_extrinsic_guess

    print("Initial extrinsic guesses")
    for k in initial_extrinsic_guesses.keys():
        print(f"Camera {k}:\n{repr(initial_extrinsic_guesses[k])}")


    # get camera intrinsics for realsense cameras.
    intrinsics = dict()
    for cam_name, cam in robot.cameras.items():
        if isinstance(cam, RealSenseCamera):
            streams = cam.rs_profile.get_streams()
            assert len(streams) == 1, "Only one stream per camera is supported at the moment and it must be the color steam. Make sure to not enable any other streams."
            color_stream = streams[0]
            color_intrinsics = color_stream.as_video_stream_profile().get_intrinsics()
            intrinsic = np.array(
                [
                    [color_intrinsics.fx, 0, color_intrinsics.ppx],
                    [0, color_intrinsics.fy, color_intrinsics.ppy],
                    [0, 0, 1],
                ]
            )
            intrinsics[cam_name] = intrinsic



    ### Data Collection Process below ###
    # We move the robot to a few joint configurations and collect images and generate a link pose dataset.

    joint_position_names = ["shoulder_pan.pos", "shoulder_lift.pos", "elbow_flex.pos", "wrist_flex.pos", "wrist_roll.pos", "gripper.pos"]
    def get_qpos(robot: SO100Follower, flat: bool = True):
        obs = robot.bus.sync_read("Present_Position")
        for k in CALIBRATION_OFFSET.keys():
            obs[k] = obs[k] - CALIBRATION_OFFSET[k]
        for k in obs.keys():
            obs[k] = np.deg2rad(obs[k])
        if not flat:
            return obs
        joint_positions = []
        for k, v in obs.items():
            joint_positions.append(v)
        joint_positions = np.array(joint_positions)
        return joint_positions
    
    def set_target_qpos(robot: SO100Follower, qpos: np.ndarray):
        action = {}
        for name, qpos_val in zip(joint_position_names, qpos):
            action[name] = np.rad2deg(qpos_val) + CALIBRATION_OFFSET[name.removesuffix(".pos")]
        robot.send_action(action)
    
    robot_def_path = Path(__file__).parent / "robot_definitions" / "so100"
    robot_urdf = URDF.load(str(robot_def_path / "so100.urdf"))

    meshes = []
    for link in robot_urdf.links:
        link_meshes = []
        for visual in link.visuals:
            link_meshes += visual.geometry.mesh.meshes
        meshes.append(merge_meshes(link_meshes))

    if args.use_previous_captures and (Path(args.output_dir) / robot_id / "link_poses_dataset.npy").exists():
        # load the previous captures
        link_poses_dataset = np.load(Path(args.output_dir) / robot_id / "link_poses_dataset.npy")
        image_dataset = np.load(Path(args.output_dir) / robot_id / "image_dataset.npy", allow_pickle=True).reshape(-1)[0]
    else:
        # reference qpos positions to calibrate with    
        qpos_samples = [
            np.array([
                0, 0, 0, np.pi / 2, np.pi / 2, 0.2
            ]),
            np.array([
                np.pi / 3, -np.pi / 6, 0, np.pi / 2, np.pi / 2, 0
            ])
        ]
        control_freq = 15
        max_radians_per_step = 0.05

        # generate our link pose dataset and image pairs. We do this by moving the robot to the reference joint positions and collecting images from all cameras
        link_poses_dataset = np.zeros((len(qpos_samples), len(meshes), 4, 4))
        image_dataset = defaultdict(list)

        for i in range(len(qpos_samples)):

            # control code for lerobot below
            goal_qpos = qpos_samples[i]
            target_qpos = get_qpos(robot)
            for _ in range(int(20*control_freq)):
                start_loop_t = time.perf_counter()
                delta_qpos = (goal_qpos - target_qpos)
                delta_step = delta_qpos.clip(
                    min=-max_radians_per_step, max=max_radians_per_step
                )
                if np.linalg.norm(delta_qpos) < 1e-4:
                    break
                target_qpos += delta_step
                dt_s = time.perf_counter() - start_loop_t
                set_target_qpos(robot, target_qpos)
                time.sleep(1 / control_freq - dt_s)
            time.sleep(1) # give some time for the robot to settle, cheap arms don't hold up as well
            qpos_dict = get_qpos(robot, flat=False)
            for cam_name, cam in robot.cameras.items():
                image_dataset[cam_name].append(cam.async_read())
                
            # get link poses
            cfg = dict()
            for k in robot_urdf.joint_map.keys():
                cfg[k] = qpos_dict[k]
            link_poses = robot_urdf.link_fk(cfg=cfg, use_names=True)
            for link_idx, v in enumerate(link_poses.values()):
                link_poses_dataset[i, link_idx] = v
        for k in image_dataset.keys():
            image_dataset[k] = np.stack(image_dataset[k])

        np.save(Path(args.output_dir) / robot_id / "link_poses_dataset.npy", link_poses_dataset)
        np.save(Path(args.output_dir) / robot_id / "image_dataset.npy", image_dataset)

    ### Camera Calibration Process below ###
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    for k in initial_extrinsic_guesses.keys():
        print(f"Calibrating camera {k}")
        initial_extrinsic_guess = initial_extrinsic_guesses[k]
        intrinsic = intrinsics[k]
        images = image_dataset[k]
        camera_mount_poses = None # TODO (stao): support this
        camera_width = images.shape[2]
        camera_height = images.shape[1]
        
        mask_path = Path(args.output_dir) / robot_id / k / f"mask.npy"
        if args.use_previous_captures and mask_path.exists():
            print(f"Using previous mask from {mask_path}")
            masks = np.load(mask_path)
        else:
            interactive_segmentation = InteractiveSegmentation(
                segmentation_model="sam2",
                segmentation_model_cfg=dict(
                    checkpoint=args.checkpoint, model_cfg=args.model_cfg
                ),
            )
            masks = interactive_segmentation.get_segmentation(images)
            np.save(mask_path, masks)

        ### run the optimization given the data ###
        predicted_camera_extrinsic_opencv = (
            optimize(
                camera_intrinsic=torch.from_numpy(intrinsic).float().to(device),
                masks=torch.from_numpy(masks).float().to(device),
                link_poses_dataset=torch.from_numpy(link_poses_dataset).float().to(device),
                initial_extrinsic_guess=torch.tensor(initial_extrinsic_guess)
                .float()
                .to(device),
                meshes=meshes,
                camera_width=camera_width,
                camera_height=camera_height,
                camera_mount_poses=(
                    torch.from_numpy(camera_mount_poses).float().to(device)
                    if camera_mount_poses is not None
                    else None
                ),
                gt_camera_pose=None,
                iterations=args.train_steps,
                early_stopping_steps=args.early_stopping_steps,
            )
            .cpu()
            .numpy()
        )
        predicted_camera_extrinsic_ros = opencv2ros(predicted_camera_extrinsic_opencv)

        ### Print predicted results ###

        print(f"Predicted camera extrinsic")
        print(f"OpenCV:\n{repr(predicted_camera_extrinsic_opencv)}")
        print(f"ROS/SAPIEN/ManiSkill/Mujoco/Isaac:\n{repr(predicted_camera_extrinsic_ros)}")

        Path(args.output_dir).mkdir(parents=True, exist_ok=True)
        np.save(
            Path(args.output_dir) / robot_id / k / "camera_extrinsic_opencv.npy",
            predicted_camera_extrinsic_opencv,
        )
        np.save(
            Path(args.output_dir) / robot_id / k / "camera_extrinsic_ros.npy",
            predicted_camera_extrinsic_ros,
        )
        np.save(Path(args.output_dir) / robot_id / k / "camera_intrinsic.npy", intrinsic)

        visualization.visualize_extrinsic_results(
            images=images,
            link_poses_dataset=link_poses_dataset,
            meshes=meshes,
            intrinsic=intrinsic,
            extrinsics=np.stack(
                [initial_extrinsic_guess, predicted_camera_extrinsic_opencv]
            ),
            masks=masks,
            labels=["Initial Extrinsic Guess", "Predicted Extrinsic"],
            output_dir=str(Path(args.output_dir) / robot_id / k),
        )
        print(f"Visualizations saved to {Path(args.output_dir) / robot_id / k}")

if __name__ == "__main__":
    main(tyro.cli(SO100Args))


from dataclasses import dataclass
from typing import Optional


@dataclass
class Args:

    urdf_path: Optional[str] = None
    """path to the URDF file of the robot. Some robot specific scripts will provide a URDF file or generate meshes for you and this argument can be left as None."""
    batch_size: Optional[int] = None
    """batch size for the optimization. If none will use whole batch optimization"""
    train_steps: int = 5000
    """number of optimization steps. The default is 5000 which is usually more than enough to converge"""
    early_stopping_steps: int = 200
    """if after this many steps of optimization the loss has not improved, then optimization will stop. If this value is 0 then early stopping is disabled."""
    seed: int = 0

    ### sam2 segmentation related configs ###
    model_cfg: str = "sam2/configs/sam2.1/sam2.1_hiera_l.yaml"
    """the model config for sam2"""
    checkpoint: str = "sam2/checkpoints/sam2.1_hiera_large.pt"
    """the checkpoint for sam2"""

    output_dir: str = "results"
    """Where to save the calibration results and any visualizations"""



from dataclasses import dataclass
from pathlib import Path

import gymnasium as gym
import mani_skill.envs
import numpy as np
import sapien
import torch
import transforms3d
import tyro
from mani_skill.envs.sapien_env import BaseEnv

from easyhec.examples.sim.base import Args
from easyhec.optim.optimize import optimize
from easyhec.segmentation.interactive import InteractiveSegmentation
from easyhec.utils import visualization
from easyhec.utils.camera_conversions import opencv2ros, ros2opencv


def generate_synthetic_data(env: BaseEnv, samples: int, camera_name: str):
    if camera_name == "human_render_camera":
        render_camera = next(iter(env._human_render_cameras.values()))
    elif camera_name in env._sensors:
        render_camera = env._sensors[camera_name]
    else:
        raise ValueError(f"Camera {camera_name} not found in environment")
    mount_link = render_camera.camera.mount
    intrinsic = render_camera.camera.get_intrinsic_matrix()[0].cpu().numpy()

    # gets ground truth camera pose. If the camera is mounted then this extrinsic is in the reference frame of the mount link
    # in sapien the poses are stored in the ROS convention
    camera_pose = ros2opencv(
        render_camera.camera.get_local_pose().sp.to_transformation_matrix()
    )
    camera_height, camera_width = (
        render_camera.config.height,
        render_camera.config.width,
    )

    # find all visible links
    visible_links = [
        x
        for x in env.agent.robot.links
        if x._objs[0].entity.find_component_by_type(sapien.render.RenderBodyComponent)
        is not None
    ]

    # generate synthetic rgb images, segmentation images, and link poses of the robot in the scene.
    images = []
    segmentation_masks = []
    link_poses_dataset = []
    camera_mount_poses = []
    init_qpos = env.agent.robot.get_qpos().cpu().numpy()[0]
    for i in range(samples):
        noise = np.random.randn(init_qpos.shape[0]) * 0.2
        qpos = init_qpos.copy() + noise
        for active_joint in env.agent.robot.active_joints:
            qpos[active_joint.active_index] = np.clip(
                qpos[active_joint.active_index],
                active_joint.limits[0, 0],
                active_joint.limits[0, 1],
            )
        env.agent.robot.set_qpos(qpos)
        env.scene.update_render()
        render_camera.capture()
        data = render_camera.get_obs(rgb=True, segmentation=True)
        images.append(data["rgb"].clone())
        segmentation_masks.append(data["segmentation"][..., 0].clone())

        if mount_link is not None:
            camera_mount_poses.append(
                mount_link.pose.sp.inv().to_transformation_matrix()
            )

        link_poses = []
        for link in visible_links:
            for render_shape in (
                link._objs[0]
                .entity.find_component_by_type(sapien.render.RenderBodyComponent)
                .render_shapes
            ):
                link_visual_mesh_pose = link.pose.sp * render_shape.local_pose
                link_poses.append(link_visual_mesh_pose.to_transformation_matrix())
        link_poses_dataset.append(np.stack(link_poses))
    link_poses_dataset = np.stack(link_poses_dataset)
    if len(camera_mount_poses) > 0:
        camera_mount_poses = np.stack(camera_mount_poses)

    segmentation_images = []
    robot_masks = []
    for i in range(len(images)):
        images[i] = images[i][0].cpu().numpy()
        segmentation_masks[i] = segmentation_masks[i][0].cpu().numpy()
    for i in range(len(segmentation_masks)):
        segmentation_image = images[i].copy()
        segment_ids = []
        for link in env.agent.robot.links:
            segment_ids.append(link.per_scene_id[0])
        robot_mask = np.isin(segmentation_masks[i], segment_ids)
        segmentation_image[robot_mask] //= 4
        robot_masks.append(robot_mask)
        segmentation_images.append(segmentation_image)
    robot_masks = np.stack(robot_masks)

    mesh_paths = []
    for link in visible_links:
        # assumes there is only one render body component
        rb = link._objs[0].entity.find_component_by_type(
            sapien.render.RenderBodyComponent
        )
        if rb is None:
            continue
        for render_shape in rb.render_shapes:
            mesh_filename = render_shape.filename
            mesh_paths.append(mesh_filename)

    return dict(
        images=images,
        segmentation_images=segmentation_images,
        mesh_paths=mesh_paths,
        segmentation_masks=segmentation_masks,
        robot_masks=robot_masks,
        link_poses_dataset=link_poses_dataset,
        intrinsic=intrinsic,
        camera_mount_poses=camera_mount_poses if len(camera_mount_poses) > 0 else None,
        camera_pose=camera_pose,
        camera_width=camera_width,
        camera_height=camera_height,
    )


@dataclass
class ManiSkillArgs(Args):
    output_dir: str = "results/maniskill"
    shader: str = "default"
    """Choice of shader to modify the rendering of the environment. default is a fast and cheap option, you can also try "rt" for ray-tracing option"""

    camera_name: str = "human_render_camera"
    """the name of the camera to try and calibrate the extrinsics for. ManiSkill envs have several cameras typically, one for human viewing labeled the human_render_camera and others for actual the actual cameras used for inputs to a robotics model. This script prints out all available cameras in the env and you can change the camera_name argument here to try and calibrate a different camera."""


def main(args: ManiSkillArgs):

    device = torch.device("cuda") if torch.cuda.is_available() else torch.device("cpu")
    torch.manual_seed(args.seed)
    np.random.seed(args.seed)

    ### create the sim environment ###
    env = gym.make(
        args.env_id,
        human_render_camera_configs=dict(shader_pack=args.shader),
        sensor_configs=dict(shader_pack=args.shader),
    )
    env.reset()
    base_env: BaseEnv = env.unwrapped

    # print out all available cameras
    print("Available sensors/cameras used for models in the environment:")
    print(list(base_env._sensors.keys()))
    print("Available human render cameras in the environment:")
    print(list(base_env._human_render_cameras.keys()))
    if args.camera_name == "human_render_camera":
        print("Using human render camera for visualization")
    else:
        print("Using camera: ", args.camera_name)

    ### generate the synthetic data ###
    synthetic_data = generate_synthetic_data(
        base_env, args.samples + args.test_samples, args.camera_name
    )

    if not args.use_ground_truth_segmentation:
        interactive_segmentation = InteractiveSegmentation(
            segmentation_model="sam2",
            segmentation_model_cfg=dict(
                checkpoint=args.checkpoint, model_cfg=args.model_cfg
            ),
        )
        synthetic_data["robot_masks"] = interactive_segmentation.get_segmentation(
            synthetic_data["images"][: args.samples]
        )

    # data that you would need to collect in the real world
    link_poses_dataset = synthetic_data["link_poses_dataset"]
    intrinsic = synthetic_data["intrinsic"]
    camera_width = synthetic_data["camera_width"]
    camera_height = synthetic_data["camera_height"]
    mesh_paths = synthetic_data["mesh_paths"]
    robot_masks = synthetic_data["robot_masks"]
    # specific to cameras that are not fixed
    camera_mount_poses = synthetic_data["camera_mount_poses"]

    # data used just for visualization
    images = synthetic_data["images"]

    ### generate an initial guess around the ground truth pose ###
    ground_truth_camera_pose = synthetic_data["camera_pose"]
    initial_extrinsic_guess = ground_truth_camera_pose.copy()

    # pick a random direction to perturb the camera position.
    rand_vec = np.random.randn(3)
    rand_vec = rand_vec / np.linalg.norm(rand_vec)
    rand_vec = rand_vec * args.initial_extrinsic_guess_pos_error
    initial_extrinsic_guess[:3, 3] -= rand_vec

    rand_vec = np.random.randn(3)
    rand_vec = rand_vec / np.linalg.norm(rand_vec)
    rand_vec = rand_vec * args.initial_extrinsic_guess_rot_error
    initial_extrinsic_guess[:3, :3] = (
        transforms3d.axangles.axangle2mat(
            rand_vec, np.deg2rad(args.initial_extrinsic_guess_rot_error)
        )
        @ initial_extrinsic_guess[:3, :3]
    )

    ### run the optimization given the data ###

    predicted_camera_extrinsic_opencv = (
        optimize(
            camera_intrinsic=torch.from_numpy(intrinsic).float().to(device),
            masks=torch.from_numpy(robot_masks[: args.samples])
            .float()
            .to(device),
            link_poses_dataset=torch.from_numpy(link_poses_dataset[: args.samples])
            .float()
            .to(device),
            initial_extrinsic_guess=torch.from_numpy(initial_extrinsic_guess)
            .float()
            .to(device),
            meshes=mesh_paths,
            camera_width=camera_width,
            camera_height=camera_height,
            camera_mount_poses=(
                torch.from_numpy(camera_mount_poses[: args.samples]).float().to(device)
                if camera_mount_poses is not None
                else None
            ),
            gt_camera_pose=torch.from_numpy(ground_truth_camera_pose)
            .float()
            .to(device),
            iterations=args.train_steps,
            early_stopping_steps=args.early_stopping_steps,
        )
        .cpu()
        .numpy()
    )
    predicted_camera_extrinsic_ros = opencv2ros(predicted_camera_extrinsic_opencv)

    ### Print predicted results ###
    print(f"Predicted camera extrinsic")
    print(f"OpenCV:\n{repr(predicted_camera_extrinsic_opencv)}")
    print(f"ROS/SAPIEN/ManiSkill/Mujoco/Isaac:\n{repr(predicted_camera_extrinsic_ros)}")

    Path(args.output_dir).mkdir(parents=True, exist_ok=True)
    np.save(
        Path(args.output_dir) / "camera_extrinsic_opencv.npy",
        predicted_camera_extrinsic_opencv,
    )
    np.save(
        Path(args.output_dir) / "camera_extrinsic_ros.npy",
        predicted_camera_extrinsic_ros,
    )
    np.save(Path(args.output_dir) / "camera_intrinsic.npy", intrinsic)

    visualization.visualize_extrinsic_results(
        images=images[args.samples :],
        link_poses_dataset=link_poses_dataset[args.samples :],
        meshes=mesh_paths,
        intrinsic=intrinsic,
        extrinsics=np.stack(
            [
                initial_extrinsic_guess,
                predicted_camera_extrinsic_opencv,
                ground_truth_camera_pose,
            ]
        ),
        camera_mount_poses=camera_mount_poses[args.samples :] if camera_mount_poses is not None else None,
        labels=["Initial Extrinsic Guess", "Predicted Extrinsic", "Ground Truth"],
        output_dir=args.output_dir,
    )
    print(f"Visualizations saved to {args.output_dir}")


if __name__ == "__main__":
    main(tyro.cli(ManiSkillArgs))



# Copyright (c) Meta Platforms, Inc. and affiliates.
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.

from typing import Tuple

import torch

from easyhec.utils.pytorch3d_so3 import hat, so3_log_map


def _so3_exp_map(
    log_rot: torch.Tensor, eps: float = 0.0001
) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor]:
    """
    A helper function that computes the so3 exponential map and,
    apart from the rotation matrix, also returns intermediate variables
    that can be re-used in other functions.
    """
    _, dim = log_rot.shape
    if dim != 3:
        raise ValueError("Input tensor shape has to be Nx3.")

    nrms = (log_rot * log_rot).sum(1)
    # phis ... rotation angles
    rot_angles = torch.clamp(nrms, eps).sqrt()
    # pyre-fixme[58]: `/` is not supported for operand types `float` and `Tensor`.
    rot_angles_inv = 1.0 / rot_angles
    fac1 = rot_angles_inv * rot_angles.sin()
    fac2 = rot_angles_inv * rot_angles_inv * (1.0 - rot_angles.cos())
    skews = hat(log_rot)
    skews_square = torch.bmm(skews, skews)

    R = (
        fac1[:, None, None] * skews
        # pyre-fixme[16]: `float` has no attribute `__getitem__`.
        + fac2[:, None, None] * skews_square
        + torch.eye(3, dtype=log_rot.dtype, device=log_rot.device)[None]
    )

    return R, rot_angles, skews, skews_square


def se3_exp_map(log_transform: torch.Tensor, eps: float = 1e-4) -> torch.Tensor:
    """
    Convert a batch of logarithmic representations of SE(3) matrices `log_transform`
    to a batch of 4x4 SE(3) matrices using the exponential map.
    See e.g. [1], Sec 9.4.2. for more detailed description.

    A SE(3) matrix has the following form:
        ```
        [ R 0 ]
        [ T 1 ] ,
        ```
    where `R` is a 3x3 rotation matrix and `T` is a 3-D translation vector.
    SE(3) matrices are commonly used to represent rigid motions or camera extrinsics.

    In the SE(3) logarithmic representation SE(3) matrices are
    represented as 6-dimensional vectors `[log_translation | log_rotation]`,
    i.e. a concatenation of two 3D vectors `log_translation` and `log_rotation`.

    The conversion from the 6D representation to a 4x4 SE(3) matrix `transform`
    is done as follows:
        ```
        transform = exp( [ hat(log_rotation) 0 ]
                         [   log_translation 1 ] ) ,
        ```
    where `exp` is the matrix exponential and `hat` is the Hat operator [2].

    Note that for any `log_transform` with `0 <= ||log_rotation|| < 2pi`
    (i.e. the rotation angle is between 0 and 2pi), the following identity holds:
    ```
    se3_log_map(se3_exponential_map(log_transform)) == log_transform
    ```

    The conversion has a singularity around `||log(transform)|| = 0`
    which is handled by clamping controlled with the `eps` argument.

    Args:
        log_transform: Batch of vectors of shape `(minibatch, 6)`.
        eps: A threshold for clipping the squared norm of the rotation logarithm
            to avoid unstable gradients in the singular case.

    Returns:
        Batch of transformation matrices of shape `(minibatch, 4, 4)`.

    Raises:
        ValueError if `log_transform` is of incorrect shape.

    [1] https://jinyongjeong.github.io/Download/SE3/jlblanco2010geometry3d_techrep.pdf
    [2] https://en.wikipedia.org/wiki/Hat_operator
    """

    if log_transform.ndim != 2 or log_transform.shape[1] != 6:
        raise ValueError("Expected input to be of shape (N, 6).")

    N, _ = log_transform.shape

    log_translation = log_transform[..., :3]
    log_rotation = log_transform[..., 3:]

    # rotation is an exponential map of log_rotation
    (
        R,
        rotation_angles,
        log_rotation_hat,
        log_rotation_hat_square,
    ) = _so3_exp_map(log_rotation, eps=eps)

    # translation is V @ T
    V = _se3_V_matrix(
        log_rotation,
        log_rotation_hat,
        log_rotation_hat_square,
        rotation_angles,
        eps=eps,
    )
    T = torch.bmm(V, log_translation[:, :, None])[:, :, 0]

    transform = torch.zeros(
        N, 4, 4, dtype=log_transform.dtype, device=log_transform.device
    )

    transform[:, :3, :3] = R
    transform[:, :3, 3] = T
    transform[:, 3, 3] = 1.0

    return transform.permute(0, 2, 1)


def se3_log_map(
    transform: torch.Tensor, eps: float = 1e-4, cos_bound: float = 1e-4
) -> torch.Tensor:
    """
    Convert a batch of 4x4 transformation matrices `transform`
    to a batch of 6-dimensional SE(3) logarithms of the SE(3) matrices.
    See e.g. [1], Sec 9.4.2. for more detailed description.

    A SE(3) matrix has the following form:
        ```
        [ R 0 ]
        [ T 1 ] ,
        ```
    where `R` is an orthonormal 3x3 rotation matrix and `T` is a 3-D translation vector.
    SE(3) matrices are commonly used to represent rigid motions or camera extrinsics.

    In the SE(3) logarithmic representation SE(3) matrices are
    represented as 6-dimensional vectors `[log_translation | log_rotation]`,
    i.e. a concatenation of two 3D vectors `log_translation` and `log_rotation`.

    The conversion from the 4x4 SE(3) matrix `transform` to the
    6D representation `log_transform = [log_translation | log_rotation]`
    is done as follows:
        ```
        log_transform = log(transform)
        log_translation = log_transform[3, :3]
        log_rotation = inv_hat(log_transform[:3, :3])
        ```
    where `log` is the matrix logarithm
    and `inv_hat` is the inverse of the Hat operator [2].

    Note that for any valid 4x4 `transform` matrix, the following identity holds:
    ```
    se3_exp_map(se3_log_map(transform)) == transform
    ```

    The conversion has a singularity around `(transform=I)` which is handled
    by clamping controlled with the `eps` and `cos_bound` arguments.

    Args:
        transform: batch of SE(3) matrices of shape `(minibatch, 4, 4)`.
        eps: A threshold for clipping the squared norm of the rotation logarithm
            to avoid division by zero in the singular case.
        cos_bound: Clamps the cosine of the rotation angle to
            [-1 + cos_bound, 3 - cos_bound] to avoid non-finite outputs.
            The non-finite outputs can be caused by passing small rotation angles
            to the `acos` function in `so3_rotation_angle` of `so3_log_map`.

    Returns:
        Batch of logarithms of input SE(3) matrices
        of shape `(minibatch, 6)`.

    Raises:
        ValueError if `transform` is of incorrect shape.
        ValueError if `R` has an unexpected trace.

    [1] https://jinyongjeong.github.io/Download/SE3/jlblanco2010geometry3d_techrep.pdf
    [2] https://en.wikipedia.org/wiki/Hat_operator
    """

    if transform.ndim != 3:
        raise ValueError("Input tensor shape has to be (N, 4, 4).")

    N, dim1, dim2 = transform.shape
    if dim1 != 4 or dim2 != 4:
        raise ValueError("Input tensor shape has to be (N, 4, 4).")

    if not torch.allclose(transform[:, :3, 3], torch.zeros_like(transform[:, :3, 3])):
        raise ValueError("All elements of `transform[:, :3, 3]` should be 0.")

    # log_rot is just so3_log_map of the upper left 3x3 block
    R = transform[:, :3, :3].permute(0, 2, 1)
    log_rotation = so3_log_map(R, eps=eps, cos_bound=cos_bound)

    # log_translation is V^-1 @ T
    T = transform[:, 3, :3]
    V = _se3_V_matrix(*_get_se3_V_input(log_rotation), eps=eps)
    log_translation = torch.linalg.solve(V, T[:, :, None])[:, :, 0]

    return torch.cat((log_translation, log_rotation), dim=1)


def _se3_V_matrix(
    log_rotation: torch.Tensor,
    log_rotation_hat: torch.Tensor,
    log_rotation_hat_square: torch.Tensor,
    rotation_angles: torch.Tensor,
    eps: float = 1e-4,
) -> torch.Tensor:
    """
    A helper function that computes the "V" matrix from [1], Sec 9.4.2.
    [1] https://jinyongjeong.github.io/Download/SE3/jlblanco2010geometry3d_techrep.pdf
    """

    V = (
        torch.eye(3, dtype=log_rotation.dtype, device=log_rotation.device)[None]
        + log_rotation_hat
        # pyre-fixme[58]: `**` is not supported for operand types `Tensor` and `int`.
        * ((1 - torch.cos(rotation_angles)) / (rotation_angles**2))[:, None, None]
        + (
            log_rotation_hat_square
            # pyre-fixme[58]: `**` is not supported for operand types `Tensor` and
            #  `int`.
            * ((rotation_angles - torch.sin(rotation_angles)) / (rotation_angles**3))[
                :, None, None
            ]
        )
    )

    return V


def _get_se3_V_input(log_rotation: torch.Tensor, eps: float = 1e-4):
    """
    A helper function that computes the input variables to the `_se3_V_matrix`
    function.
    """
    # pyre-fixme[58]: `**` is not supported for operand types `Tensor` and `int`.
    nrms = (log_rotation**2).sum(-1)
    rotation_angles = torch.clamp(nrms, eps).sqrt()
    log_rotation_hat = hat(log_rotation)
    log_rotation_hat_square = torch.bmm(log_rotation_hat, log_rotation_hat)
    return log_rotation, log_rotation_hat, log_rotation_hat_square, rotation_angles



try:
    import sam2
except ModuleNotFoundError:
    print("sam2 is not installed. Features using SAM2 for segmentation will not work. Follow the instructions at https://github.com/facebookresearch/sam2 to install it.")

try:
    import nvdiffrast
except ModuleNotFoundError:
    print("nvdiffrast is not installed. Please install it with pip install 'nvdiffrast @ git+https://github.com/NVlabs/nvdiffrast.git@729261dc64c4241ea36efda84fbf532cc8b425b8'")
    exit(1)


import os.path as osp
from pathlib import Path
from typing import List, Optional

import cv2
import matplotlib.pyplot as plt
import numpy as np
import torch
import trimesh
from tqdm import tqdm

from easyhec.optim.nvdiffrast_renderer import NVDiffrastRenderer


def visualize_extrinsic_results(
    images,
    link_poses_dataset,
    meshes,
    intrinsic: np.ndarray,
    extrinsics: np.ndarray,
    camera_mount_poses: Optional[np.ndarray] = None,
    masks: Optional[np.ndarray] = None,
    labels: List[str] = [],
    output_dir="results/",
    return_rgb: bool = False,
):
    """
    Visualizes a given list of extrinsic matrices and draws the mask cameras at those extrinsics would project on the original RGB images.

    Args:
        images (np.ndarray, shape (N, H, W, 3)): List of RGB images to visualize.
        link_poses_dataset (np.ndarray, shape (N, L, 4, 4)): Link poses relative to any frame (e.g. the robot base frame), where N is the number of samples, L is the number of links
        meshes (List[str | trimesh.Trimesh]): List of mesh paths or trimesh.Trimesh objects for each of the links
        intrinsic (np.ndarray, shape (3, 3)): Camera intrinsic matrix
        extrinsics (np.ndarray, shape (M, 4, 4)): Extrinsic matrices to visualize
        camera_mount_poses (np.ndarray, shape (N, 4, 4)): Camera mount poses relative to the robot base frame, where N is the number of samples. If none then camera is assumed to be fixed.
        masks (np.ndarray, shape (N, H, W)): If given, will also display an image showing the masks used for optimization on top of the original images.
        labels (List[str]): List of labels for each of the extrinsics
        output_dir (str): Directory to save the visualizations
    """
    ### visualization code for the predicted extrinsic ###
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    camera_height, camera_width = images[0].shape[:2]
    renderer = NVDiffrastRenderer(camera_height, camera_width)
    extrinsics = torch.from_numpy(extrinsics).float().to(device)
    if camera_mount_poses is not None:
        camera_mount_poses = torch.from_numpy(camera_mount_poses).float().to(device)
    link_poses_dataset = torch.from_numpy(link_poses_dataset).float().to(device)

    for i in range(len(meshes)):
        if isinstance(meshes[i], str):
            meshes[i] = trimesh.load(osp.expanduser(meshes[i]), force="mesh")
    link_vertices = [mesh.vertices.copy() for mesh in meshes]
    link_faces = [mesh.faces.copy() for mesh in meshes]
    link_vertices = [
        torch.from_numpy(mesh.vertices).float().to(device) for mesh in meshes
    ]
    link_faces = [torch.from_numpy(mesh.faces).int().to(device) for mesh in meshes]

    def get_mask_from_camera_pose(camera_pose):
        mask = torch.zeros((camera_height, camera_width), device=device)
        for j, link_pose in enumerate(link_poses_dataset[i]):
            link_mask = renderer.render_mask(
                link_vertices[j],
                link_faces[j],
                intrinsic,
                camera_pose @ link_pose,
            )
            link_mask = link_mask.detach()
            mask[link_mask > 0] = 1
        return mask

    for i in tqdm(range(len(images))):
        overlaid_images = []
        for j in range(len(extrinsics)):
            if camera_mount_poses is not None:
                mask = get_mask_from_camera_pose(extrinsics[j] @ camera_mount_poses[i])
            else:
                mask = get_mask_from_camera_pose(extrinsics[j])
            mask = mask.cpu().numpy()
            overlaid_images.append(images[i].copy())
            overlaid_images[-1][mask > 0] = overlaid_images[-1][mask > 0] // 4

        num_subplots = len(extrinsics) + 1 if masks is not None else len(extrinsics)

        plt.rcParams.update({'font.size': 16})  # Increase font size for all text elements
        fig = plt.figure(figsize=(7 * num_subplots, 8))
        for j in range(len(extrinsics)):
            ax = fig.add_subplot(1, num_subplots, j + 1)
            ax.imshow(overlaid_images[j])
            ax.axis("off")
            ax.set_title(labels[j])
        
        if masks is not None:
            ax = fig.add_subplot(1, num_subplots, num_subplots)
            reference_mask = images[i].copy()
            reference_mask[masks[i] > 0] = reference_mask[masks[i] > 0] // 4
            ax.imshow(reference_mask)
            ax.axis("off")
            ax.set_title("Masks")

        Path(output_dir).mkdir(parents=True, exist_ok=True)
        plt.tight_layout()
        fig.savefig(f"{output_dir}/{i}.png")
        plt.close()
        if return_rgb:
            return cv2.cvtColor(cv2.imread(f"{output_dir}/{i}.png"), cv2.COLOR_BGR2RGB)


# Copyright (c) Meta Platforms, Inc. and affiliates.
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.

import torch


def hat(v: torch.Tensor) -> torch.Tensor:
    """
    Compute the Hat operator [1] of a batch of 3D vectors.

    Args:
        v: Batch of vectors of shape `(minibatch , 3)`.

    Returns:
        Batch of skew-symmetric matrices of shape
        `(minibatch, 3 , 3)` where each matrix is of the form:
            `[    0  -v_z   v_y ]
             [  v_z     0  -v_x ]
             [ -v_y   v_x     0 ]`

    Raises:
        ValueError if `v` is of incorrect shape.

    [1] https://en.wikipedia.org/wiki/Hat_operator
    """

    N, dim = v.shape
    if dim != 3:
        raise ValueError("Input vectors have to be 3-dimensional.")

    h = torch.zeros((N, 3, 3), dtype=v.dtype, device=v.device)

    x, y, z = v.unbind(1)

    h[:, 0, 1] = -z
    h[:, 0, 2] = y
    h[:, 1, 0] = z
    h[:, 1, 2] = -x
    h[:, 2, 0] = -y
    h[:, 2, 1] = x

    return h


def matrix_to_axis_angle(matrix: torch.Tensor, fast: bool = False) -> torch.Tensor:
    """
    Convert rotations given as rotation matrices to axis/angle.

    Args:
        matrix: Rotation matrices as tensor of shape (..., 3, 3).
        fast: Whether to use the new faster implementation (based on the
            Rodrigues formula) instead of the original implementation (which
            first converted to a quaternion and then back to a rotation matrix).

    Returns:
        Rotations given as a vector in axis angle form, as a tensor
            of shape (..., 3), where the magnitude is the angle
            turned anticlockwise in radians around the vector's
            direction.

    """
    if not fast:
        return quaternion_to_axis_angle(matrix_to_quaternion(matrix))

    if matrix.size(-1) != 3 or matrix.size(-2) != 3:
        raise ValueError(f"Invalid rotation matrix shape {matrix.shape}.")

    omegas = torch.stack(
        [
            matrix[..., 2, 1] - matrix[..., 1, 2],
            matrix[..., 0, 2] - matrix[..., 2, 0],
            matrix[..., 1, 0] - matrix[..., 0, 1],
        ],
        dim=-1,
    )
    norms = torch.norm(omegas, p=2, dim=-1, keepdim=True)
    traces = torch.diagonal(matrix, dim1=-2, dim2=-1).sum(-1).unsqueeze(-1)
    angles = torch.atan2(norms, traces - 1)

    zeros = torch.zeros(3, dtype=matrix.dtype, device=matrix.device)
    omegas = torch.where(torch.isclose(angles, torch.zeros_like(angles)), zeros, omegas)

    near_pi = angles.isclose(angles.new_full((1,), torch.pi)).squeeze(-1)

    axis_angles = torch.empty_like(omegas)
    axis_angles[~near_pi] = (
        0.5 * omegas[~near_pi] / torch.sinc(angles[~near_pi] / torch.pi)
    )

    # this derives from: nnT = (R + 1) / 2
    n = 0.5 * (
        matrix[near_pi][..., 0, :]
        + torch.eye(1, 3, dtype=matrix.dtype, device=matrix.device)
    )
    axis_angles[near_pi] = angles[near_pi] * n / torch.norm(n)

    return axis_angles


def so3_log_map(
    R: torch.Tensor, eps: float = 0.0001, cos_bound: float = 1e-4
) -> torch.Tensor:
    """
    Convert a batch of 3x3 rotation matrices `R`
    to a batch of 3-dimensional matrix logarithms of rotation matrices
    The conversion has a singularity around `(R=I)`.

    Args:
        R: batch of rotation matrices of shape `(minibatch, 3, 3)`.
        eps: (unused, for backward compatibility)
        cos_bound: (unused, for backward compatibility)

    Returns:
        Batch of logarithms of input rotation matrices
        of shape `(minibatch, 3)`.
    """

    N, dim1, dim2 = R.shape
    if dim1 != 3 or dim2 != 3:
        raise ValueError("Input has to be a batch of 3x3 Tensors.")

    return matrix_to_axis_angle(R)



from dataclasses import dataclass
from typing import Annotated, Optional

import tyro


@dataclass
class Args:
    env_id: Annotated[str, tyro.conf.arg(aliases=["-e"])] = "StackCube-v1"
    """the simulated environment ID if you want to test easy hec on a sim environment. With the sim environment it will use ground truth segmentation masks for optimization"""
    samples: int = 10
    """number of synthetic samples to generate to try and calibrate the camera with"""
    test_samples: int = 10
    """number of synthetic samples to generate to test the predicted extrinsics with"""
    initial_extrinsic_guess_pos_error: float = 0.1
    """error in the initial extrinsic translation guess in meters."""
    initial_extrinsic_guess_rot_error: float = 15.0
    """error in the initial extrinsic rotation guess in degrees."""
    batch_size: Optional[int] = None
    """batch size for the optimization. If none will use whole batch optimization"""
    train_steps: int = 5000
    """number of optimization steps. The default is 5000 which is usually more than enough to converge"""
    early_stopping_steps: int = 200
    """if after this many steps of optimization the loss has not improved, then optimization will stop. If this value is 0 then early stopping is disabled."""
    seed: int = 0

    ### sam2 segmentation related configs ###
    use_ground_truth_segmentation: bool = True
    """if true will optimize against the ground truth segmentation masks. If false will open a GUI with an interactive segmentation tool that a user can click points on that are then sent to the SAM2 segmentation model to predict segmentation masks"""
    model_cfg: str = "sam2/configs/sam2.1/sam2.1_hiera_l.yaml"
    """the model config for sam2"""
    checkpoint: str = "sam2/checkpoints/sam2.1_hiera_large.pt"
    """the checkpoint for sam2"""

    output_dir: str = "results/sim"
    """Where to save the calibration results and any visualizations"""



from dataclasses import dataclass
from pathlib import Path

import cv2
import numpy as np
import pyrealsense2 as rs
import torch
import trimesh
import tyro
from transforms3d.euler import euler2mat

from easyhec.examples.real.base import Args
from easyhec.optim.optimize import optimize
from easyhec.segmentation.interactive import InteractiveSegmentation
from easyhec.utils import visualization
from easyhec.utils.camera_conversions import opencv2ros, ros2opencv


@dataclass
class RealPaperArgs(Args):
    """Calibrate a (realsense) camera with just a piece of standard sized paper. Note that this script might not work with your particular realsense camera, modify as needed.Other cameras can work if you modify the code to get the camera intrinsics and a single color image from the camera."""
    output_dir: str = "results/paper"
    paper_type: str = "letter"
    """The type of paper to use to calibrate against. Options are 'letter' or 'a4'"""
    realsense_camera_serial_id: str = "none"
    """The serial id of the realsense camera to use for calibration"""
    # TODO (stao): A1, A2, A3, follow a nice structure, we can just generate the meshes for those.


paper_sizes = {
    "letter": {
        "width": 0.2159,  # 8.5 inches in mm
        "height": 0.2794,  # 11 inches in mm
    },
    "a4": {
        "width": 0.210,  # 8.27 inches in mm
        "height": 0.297,  # 11.69 inches in mm
    },
}


def main(args: RealPaperArgs):

    device = torch.device("cuda") if torch.cuda.is_available() else torch.device("cpu")
    torch.manual_seed(args.seed)
    np.random.seed(args.seed)

    # Initialize RealSense configuration
    config = rs.config()
    pipeline = rs.pipeline()
    ctx = rs.context()
    devices = ctx.query_devices()
    if len(devices) == 0:
        raise RuntimeError("No RealSense devices found.")

    # Configure streams
    camera_width = 1280
    camera_height = 720
    if args.realsense_camera_serial_id == "none":
        print("No realsense camera serial id provided, using the first device found")
        realsense_camera_serial_id = devices[0].get_info(rs.camera_info.serial_number)
    else:
        realsense_camera_serial_id = args.realsense_camera_serial_id
    print(f"RealSense device id: {realsense_camera_serial_id}")
    config.enable_device(realsense_camera_serial_id)
    config.enable_stream(
        rs.stream.color, camera_width, camera_height, rs.format.bgr8, 30
    )
    # Get the color stream profile and its intrinsics
    profile = pipeline.start(config)
    color_stream = profile.get_stream(rs.stream.color)

    ### Fetch Intrinsics ###
    color_intrinsics = color_stream.as_video_stream_profile().get_intrinsics()
    intrinsic = np.array(
        [
            [color_intrinsics.fx, 0, color_intrinsics.ppx],
            [0, color_intrinsics.fy, color_intrinsics.ppy],
            [0, 0, 1],
        ],
        dtype=np.float32,
    )

    ### Fetch one color image ###
    skip_frames = 60
    print("Starting camera and warming it up...")
    for _ in range(skip_frames):
        frames = pipeline.wait_for_frames()
        cframe = frames.get_color_frame()
        if not cframe:
            print("No frame")
            continue
        image = np.asanyarray(cframe.get_data())
        image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)

    print(f"Camera Intrinsics:\n {repr(intrinsic)}")
    images = [image]

    ### Make an initial guess for the extrinsic ###
    # use what we put in sim as the initial guess
    initial_extrinsic_guess = np.eye(4)

    # the guess says we are at position xyz=[-0.4, 0.0, 0.4] and angle the camerea downwards by np.pi / 4 radians  or 45 degrees
    # note that this convention is more natural for robotics (follows the typical convention for ROS and various simulators), where +Z is moving up towards the sky, +Y is to the left, +X is forward
    initial_extrinsic_guess[:3, :3] = euler2mat(0, np.pi / 4, 0)
    initial_extrinsic_guess[:3, 3] = np.array([-0.4, 0.1, 0.4])
    initial_extrinsic_guess = ros2opencv(initial_extrinsic_guess)

    print("Initial extrinsic guess", initial_extrinsic_guess)

    # Create a box mesh representing the letter paper (in meters)
    paper_width = paper_sizes[args.paper_type]["width"]
    paper_height = paper_sizes[args.paper_type]["height"]
    paper_box = trimesh.creation.box(extents=(paper_width, paper_height, 1e-3))
    meshes = [paper_box]
    # We assume the world frame is centered at the paper and oriented to be perpendicular to the paper
    link_poses_dataset = np.stack(np.eye(4)).reshape(1, 1, 4, 4)

    camera_mount_poses = None

    interactive_segmentation = InteractiveSegmentation(
        segmentation_model="sam2",
        segmentation_model_cfg=dict(
            checkpoint=args.checkpoint, model_cfg=args.model_cfg
        ),
    )
    masks = interactive_segmentation.get_segmentation(images)

    ### run the optimization given the data ###
    predicted_camera_extrinsic_opencv = (
        optimize(
            camera_intrinsic=torch.from_numpy(intrinsic).float().to(device),
            masks=torch.from_numpy(masks).float().to(device),
            link_poses_dataset=torch.from_numpy(link_poses_dataset).float().to(device),
            initial_extrinsic_guess=torch.tensor(initial_extrinsic_guess)
            .float()
            .to(device),
            meshes=meshes,
            camera_width=camera_width,
            camera_height=camera_height,
            camera_mount_poses=(
                torch.from_numpy(camera_mount_poses).float().to(device)
                if camera_mount_poses is not None
                else None
            ),
            gt_camera_pose=None,
            iterations=args.train_steps,
            early_stopping_steps=args.early_stopping_steps,
        )
        .cpu()
        .numpy()
    )
    predicted_camera_extrinsic_ros = opencv2ros(predicted_camera_extrinsic_opencv)

    ### Print predicted results ###

    print(f"Predicted camera extrinsic")
    print(f"OpenCV:\n{repr(predicted_camera_extrinsic_opencv)}")
    print(f"ROS/SAPIEN/ManiSkill/Mujoco/Isaac:\n{repr(predicted_camera_extrinsic_ros)}")

    Path(args.output_dir).mkdir(parents=True, exist_ok=True)
    np.save(
        Path(args.output_dir) / "camera_extrinsic_opencv.npy",
        predicted_camera_extrinsic_opencv,
    )
    np.save(
        Path(args.output_dir) / "camera_extrinsic_ros.npy",
        predicted_camera_extrinsic_ros,
    )
    np.save(Path(args.output_dir) / "camera_intrinsic.npy", intrinsic)

    visualization.visualize_extrinsic_results(
        images=images,
        link_poses_dataset=link_poses_dataset,
        meshes=meshes,
        intrinsic=intrinsic,
        extrinsics=np.stack(
            [initial_extrinsic_guess, predicted_camera_extrinsic_opencv]
        ),
        masks=masks,
        labels=["Initial Extrinsic Guess", "Predicted Extrinsic"],
        output_dir=args.output_dir,
    )
    print(f"Visualizations saved to {args.output_dir}")


if __name__ == "__main__":
    main(tyro.cli(RealPaperArgs))



from typing import List, Optional

import torch
from tqdm import tqdm

from easyhec.optim.rb_solver import RBSolver, RBSolverConfig


def optimize(
    initial_extrinsic_guess: torch.Tensor,
    camera_intrinsic: torch.Tensor,
    masks: torch.Tensor,
    link_poses_dataset: torch.Tensor,
    meshes: List[str],
    camera_width: int,
    camera_height: int,
    camera_mount_poses: Optional[torch.Tensor] = None,
    iterations: int = 5000,
    learning_rate: float = 3e-3,
    gt_camera_pose: Optional[torch.Tensor] = None,
    batch_size: Optional[int] = None,
    early_stopping_steps: int = 200,
    verbose: bool = True,
    return_history: bool = False,
):
    """
    Optimizes an initial guess of a camera extrinsic using the camera intrinsic matrix, a dataset of robot masks, link poses relative to the robot base frame, and paths to the mesh files of each of the link poses.

    Inputs are expected to be torch tensors on the same device. If they are on the GPU, all optimization will be done on the GPU. Poses are expected to follow the opencv2 transformation conventions.


    Parameters:

        initial_extrinsic_guess (torch.Tensor, shape (4, 4)): Initial guess of the camera extrinsic
        camera_intrinsic (torch.Tensor, shape (3, 3)): Camera intrinsic matrix
        masks (torch.Tensor, shape (N, H, W)): Robot segmentation masks
        link_poses_dataset (torch.Tensor, shape (N, L, 4, 4)): Link poses relative to any frame (e.g. the robot base frame), where N is the number of samples, L is the number of links
        meshes (List[str | trimesh.Trimesh]): List of paths to the mesh files of each of the L links. Can also be a list of trimesh.Trimesh objects.
        camera_width (int): Camera width
        camera_height (int): Camera height
        camera_mount_poses (torch.Tensor, shape (N, 4, 4)): Used for cameras that are fixed relative to some mount that may be moving. If None, then the camera is assumed to be fixed. If provided the initial extrinsic guess should be relative to the mount frame.
        iterations (int): Number of optimization iterations
        learning_rate (float): Learning rate for the Adam optimizer
        batch_size (int): Default is None meaning whole batch optimization. Otherwise this specifies the number of samples to process in each batch.
        gt_camera_pose (torch.Tensor, shape (4, 4)): Default is None. If a ground truth camera pose is provided the optimization function will compute error metrics relative to the ground truth camera pose.
        early_stopping_steps (int): Default is 200. If the loss has not improved after this many steps the optimization will stop.
        verbose (bool): Default is True. If True, will print the loss value and a progress bar.
        return_history (bool): Default is False. If True, will return a list of all the current and previous best predicted extrinsics.
    """
    device = initial_extrinsic_guess.device
    cfg = RBSolverConfig(
        camera_width=camera_width,
        camera_height=camera_height,
        robot_masks=masks,
        link_poses_dataset=link_poses_dataset,
        meshes=meshes,
        initial_extrinsic_guess=initial_extrinsic_guess,
    )
    solver = RBSolver(cfg)
    solver = solver.to(device)
    optimizer = torch.optim.Adam(solver.parameters(), lr=learning_rate)
    best_predicted_extrinsic = initial_extrinsic_guess.clone()
    best_loss = float("inf")
    last_loss_improvement_step = 0
    pbar = tqdm(range(iterations)) if verbose else range(iterations)
    dataset = dict(
        intrinsic=camera_intrinsic,
        link_poses=link_poses_dataset,
        mask=masks,
        mount_poses=camera_mount_poses,
    )
    if gt_camera_pose is not None:
        dataset["gt_camera_pose"] = gt_camera_pose

    if return_history:
        extrinsics = []
    for i in pbar:
        if batch_size is None:
            batch = dataset
        else:
            bid = torch.randperm(len(dataset["mask"]))[:batch_size]
            batch = {k: v[bid] for k, v in dataset.items()}
        output = solver(batch)
        optimizer.zero_grad()
        output["mask_loss"].backward()
        optimizer.step()
        loss_value = output["mask_loss"].item()
        if loss_value < best_loss:
            best_loss = loss_value
            best_predicted_extrinsic = solver.get_predicted_extrinsic()
            last_loss_improvement_step = i
            if return_history:
                extrinsics.append(best_predicted_extrinsic)
        if i - last_loss_improvement_step >= early_stopping_steps:
            break
        if verbose:
            pbar.set_description(f"Loss: {loss_value:.2f}, Best Loss: {best_loss:.2f}")
        if "metrics" in output:
            pbar.set_postfix(output["metrics"])
    if return_history:
        return torch.stack(extrinsics)
    else:
        return best_predicted_extrinsic



from setuptools import find_packages, setup

setup(
    name="easyhec",
    version="0.1.4",
    packages=find_packages(),
    author="Stone Tao",
    homepage="https://github.com/stonet2000/easyhec",
    summary="EasyHec is a library for fast and automatic camera extrinsic calibration",
    license="BSD-3-Clause",
    url="https://github.com/stonet2000/easyhec",
    python_requires=">=3.9",
    install_requires=[
        "tyro",
        "torch",
        "tqdm",
        "opencv-python",
        "trimesh",
        "transforms3d",
        "matplotlib",
        # ninja is used by nvdiffrast
        "ninja>=1.11",
    ],
    extras_require={
        "dev": [
            "black",
            "isort",
            "pre-commit",
        ],
        "sim-maniskill": [
            "mani_skill-nightly",
        ],
    },
)



import os.path as osp
from dataclasses import dataclass
from typing import List, Union

import numpy as np
import torch
import torch.nn as nn
import trimesh

from easyhec.optim.nvdiffrast_renderer import NVDiffrastRenderer
from easyhec.utils import utils_3d


@dataclass
class RBSolverConfig:
    camera_height: int
    camera_width: int
    robot_masks: torch.Tensor
    link_poses_dataset: torch.Tensor
    meshes: List[Union[str, trimesh.Trimesh]]
    initial_extrinsic_guess: torch.Tensor


class RBSolver(nn.Module):
    """Rendering based solver for inverse rendering based extrinsic prediction"""

    def __init__(self, cfg: RBSolverConfig):
        super().__init__()
        self.cfg = cfg
        meshes = self.cfg.meshes
        for link_idx, mesh in enumerate(meshes):
            if isinstance(mesh, str):
                mesh = trimesh.load(osp.expanduser(mesh), force="mesh")
            else:
                assert isinstance(mesh, trimesh.Trimesh)
            vertices = torch.from_numpy(mesh.vertices).float()
            faces = torch.from_numpy(mesh.faces).int()
            self.register_buffer(f"vertices_{link_idx}", vertices)
            self.register_buffer(f"faces_{link_idx}", faces)
        self.nlinks = len(meshes)
        # camera parameters
        init_Tc_c2b = self.cfg.initial_extrinsic_guess
        init_dof = utils_3d.se3_log_map(
            torch.as_tensor(init_Tc_c2b, dtype=torch.float32)[None].permute(0, 2, 1),
            eps=1e-5,
            backend="opencv",
        )[0]
        self.dof = nn.Parameter(init_dof, requires_grad=True)
        # setup renderer
        self.H, self.W = self.cfg.camera_height, self.cfg.camera_width
        self.renderer = NVDiffrastRenderer(self.H, self.W)
        self.register_buffer(f"history_ops", torch.zeros(10000, 6))

    def forward(self, data):
        put_id = (self.history_ops == 0).all(dim=1).nonzero()[0, 0].item()
        self.history_ops[put_id] = self.dof.detach()
        Tc_c2b = utils_3d.se3_exp_map(self.dof[None]).permute(0, 2, 1)[0]
        losses = []
        all_frame_all_link_si = []
        masks_ref = data["mask"]
        link_poses = data["link_poses"]
        intrinsic = data["intrinsic"]
        mount_poses = data["mount_poses"] if "mount_poses" in data else None
        assert link_poses.shape[0] == masks_ref.shape[0]
        assert link_poses.shape[1:] == (self.nlinks, 4, 4)
        assert masks_ref.shape[1:] == (self.H, self.W)

        batch_size = masks_ref.shape[0]
        for bid in range(batch_size):
            all_link_si = []
            for link_idx in range(self.nlinks):
                if mount_poses is not None:
                    Tc_c2l = Tc_c2b @ mount_poses[bid] @ link_poses[bid, link_idx]
                else:
                    Tc_c2l = Tc_c2b @ link_poses[bid, link_idx]
                verts, faces = getattr(self, f"vertices_{link_idx}"), getattr(
                    self, f"faces_{link_idx}"
                )
                si = self.renderer.render_mask(verts, faces, intrinsic, Tc_c2l)
                all_link_si.append(si)
            if len(all_link_si) == 1:
                all_link_si = all_link_si[0].reshape(1, self.H, self.W)
            else:
                all_link_si = torch.stack(all_link_si)
            all_link_si = all_link_si.sum(0).clamp(max=1)
            all_frame_all_link_si.append(all_link_si)
            loss = torch.sum((all_link_si - masks_ref[bid].float()) ** 2)
            losses.append(loss)
        loss = torch.stack(losses).mean()
        all_frame_all_link_si = torch.stack(all_frame_all_link_si)
        output = {
            "rendered_masks": all_frame_all_link_si,
            "ref_masks": masks_ref,
            "error_maps": (all_frame_all_link_si - masks_ref.float()).abs(),
        }
        # metrics
        output = dict()
        if "gt_camera_pose" in data:
            gt_Tc_c2b = data["gt_camera_pose"]
            if not torch.allclose(gt_Tc_c2b, torch.eye(4).to(gt_Tc_c2b.device)):
                gt_dof6 = utils_3d.se3_log_map(
                    gt_Tc_c2b[None].permute(0, 2, 1), backend="opencv"
                )[0]
                trans_err = ((gt_dof6[:3] - self.dof[:3]) * 100).abs()
                rot_err = (gt_dof6[3:] - self.dof[3:]).abs().max() / np.pi * 180

                metrics = {
                    "err_x": trans_err[0].item(),
                    "err_y": trans_err[1].item(),
                    "err_z": trans_err[2].item(),
                    "err_trans": trans_err.norm().item(),
                    "err_rot": rot_err.item(),
                }
                output["metrics"] = metrics
        output["mask_loss"] = loss
        return output

    def get_predicted_extrinsic(self):
        return utils_3d.se3_exp_map(self.dof[None].detach()).permute(0, 2, 1)[0]
